/*
 * Program:   R  A  C
 * 
 * RAC - Rows and Columns, the text extraction tool 
 * http://aurelio.net/rac
 *
 * Copyright 2003  Aurelio Marinho Jargas
 *                 Thobias Salazar Trevisan
 *
 * Please address all bugs and comments to:
 *    verde@aurelio.net
 *    thobias@thobias.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You have received a copy of the GNU General Public License along
 *  with this program, on the COPYING file.
 *
 */

#include "rac.h"


/* function defines the LINE_START, LINE_END and STEP_LINE (maybe) 
 * address of a command
 */
void get_line_range(char *command)
{
	char *p1, *temp_line, c, *temp_p1;
	int i, j=0;

	p1 = xmalloc(strlen(command)+1, "get_line_range alloc p1.");
	temp_p1 = p1;
	p1[0] = '\0';
	/* erase between '[' and ']' 
	 * now, p1 has the addresss without [xxxxx]s */ 
	for (i=0;i<strlen(command);i++){
		if (command[i] != '[' && command[i] != ']' && j == 0)
			strncat(p1,&command[i],1);
		if (command[i] == '[') j = 1;
		if (command[i] == ']') j = 0;
	}
	/* we have some line address or just a '[xxxx]' address ?! */ 
	if (p1[0] == '\0') {
		rac.LINE_START =  1;
		rac.LINE_END   = -1;
		free(temp_p1);
		return;
	}
	/* to record the chunks (chars) of read line address */
	temp_line = xmalloc(strlen(p1)+1, "get_line_range alloc tem_line.");
	temp_line[0] = '\0';
#ifdef DEBUG2
	printf("--- only lines command = %s\n",p1);
#endif
	strncpy(&c,p1,1);
	while (c) {
		/* test what kind of address.
		 * it has line range */
		if (c == ':') {
			rac.LINE_START = atoi(temp_line) == '\0' ? 1 : atoi(temp_line);
			temp_line[0] = '\0';
			p1++;
			strncpy(temp_line,p1,strlen(p1));
			rac.LINE_END = p1[0] == '\0' ? -1 : atoi(p1);
			break;
		}
		/* line step */
		else if (c == '~') {
			rac.LINE_START = atoi(temp_line);
			if (rac.LINE_START == 0)
				break;
				/*error("", "line start can not be zero");*/

			temp_line[0] = '\0';
			p1++;
			strncpy(&c,p1,1);
			rac.STEP_LINE = !c ? 0 : atoi(p1);
			rac.LINE_END = rac.STEP_LINE < 0 ? 1 : -1;
			if (rac.STEP_LINE == 0)
				error("", "STEP_LINE can not be zero");
		}
		strncat(temp_line,p1,1);
		p1++;
		strncpy(&c,p1,1);
		/* one line only ex: 'x' */
		if (!c && rac.LINE_START == 0 && rac.LINE_END == 0)
				rac.LINE_START = rac.LINE_END = atoi(temp_line);
	}

#ifdef DEBUG2
	printf("--- line_start = %li line_end = %li STEP_LINE %li\n",rac.LINE_START,rac.LINE_END,rac.STEP_LINE);
#endif
	free(temp_line);
	free(temp_p1);
}

/* check and adjust start and end points 
 * its the same to lines and columns 
 * it returns -1 if its a invalida range 
 */
int check_range(long int *start, long int *end, long int total, int range_adjust)
{
	/* convert the negative address to positive */
	if (*start < 0) *start = total + (*start+1);
	if (*end < 0)   *end   = total + (*end+1);

	if (range_adjust == RANGE_ADJUST) {
		if (*start <= total && *start >= 1) 
			*end = *end < 1 ? 1 : (*end > total ? total : *end);
		if (*end <= total && *end >= 1) 
			*start = *start < 1 ? 1 : (*start > total ? total : *start);
	}

	return (*start > total || *start < 1 || *end > total || *end < 1 ? -1 : 1);
}

/* executes (print) commands type = TYPE_COL_RANGE
 * ex:  4[1,-1,4:9],  [4:8],  ...
 */
void exec_col_range(char *command, char *col_cmd_start, char *col_cmd_end)
{
	char *token, *temp_col, *p, *temp_line, *temp_step_col;
	long int i, j, print_newline = 0;
	long int col_start, col_end, temp_col_check;
	
	/* check and adapt the line number */
	if (check_range(&rac.LINE_START, &rac.LINE_END, rac.total_lines, RANGE_NOT_ADJUST) < 1)
		return;

#ifdef DEBUG2
	printf("--- Definitive Line Range = from %li to %li\n",rac.LINE_START,rac.LINE_END);
	printf("--- col start %s end %s\n",col_cmd_start, col_cmd_end);
#endif

	 /* if we have more than 1 line (ex: '[1,-1]), than we walk (to break the column
	  * command) 'n' (number of lines) times in the col_cmd_start memory */
	temp_col = xmalloc(strlen(col_cmd_start)+1,"exec_col_range alloc temp_col.");

	for (i=rac.LINE_START; i<=rac.LINE_END; i++) {
		/* locate and get the line */
		locate_line(i);
		temp_line = get_line(i);
		strncpy(temp_col,col_cmd_start,strlen(col_cmd_start));
		temp_col[strlen(col_cmd_start)] = '\0';
		/* break the command into separate commands */
		token = strtok(temp_col, ",");
		while (token != NULL) {
			rac.STEP_COL = 0;
			/* the column command has a specific STEP_COL */
			if ((temp_step_col=strchr(token,'~')) != NULL)
				rac.STEP_COL = atoi(++temp_step_col);
			/* it is a single char or STEP_COL range */
			if ((p = strchr(token,':')) == NULL) {
				col_start = atoi(token);
				temp_col_check = rac.STEP_COL > 0 ? -1 : 1;
				if (check_range(&col_start, &temp_col_check, rac.num_chars_line[i-1]-1, RANGE_NOT_ADJUST) > 0) {
					if (rac.STEP_COL == 0)
						printf("%c",temp_line[col_start-1]);
					else
						for (j=col_start-1; j!=(temp_col_check-1)+rac.STEP_COL; j+=rac.STEP_COL) {
							/* as for uses != condition, we have to guarantee
							 * that we do not access away from the limits */
							if (col_start-1 < temp_col_check-1 && j > temp_col_check-1) break;
							if (col_start-1 > temp_col_check-1 && j < temp_col_check-1) break;
							printf("%c",temp_line[j]);
						}
					print_newline = 1;
				}
			}
			/* it is a column range */
			else {
				col_start = atoi(token) != 0 ? atoi(token) : 1;
				p++;
				col_end = atoi(p) != 0 ? atoi(p) : -1;
				if (check_range(&col_start, &col_end, rac.num_chars_line[i-1]-1, RANGE_ADJUST) > 0) {
					if (rac.STEP_COL == 0) rac.STEP_COL = col_start <= col_end ? 1 : -1;
					for (j=col_start-1; j!=(col_end-1)+rac.STEP_COL; j+=rac.STEP_COL) {
						printf("%c",temp_line[j]);
					}
					print_newline = 1;
				}
			}
			/* get the next command */
			token = strtok(NULL, ",");
		}
		free(temp_line);
		if (print_newline == 1) {
			printf("\n");
			print_newline = 0;
		}
	}
	free(temp_col);
}

/* it executes (print) commands type = TYPE_LINE_RANGE
 * ex: 2,  1~2, 3:4,  3:4[2],  1[2]:4[-1],  1:-1,  ...  
 */
void exec_line_range(char *command, char *col_cmd_start, char *col_cmd_end)
{
	long int i, j;
	long int col_start, col_end, temp_col=1;
	char *temp_line;

	/* just one line or STEP_LINE is defined
	 * so we can not adjust the range */
	if (rac.LINE_START == rac.LINE_END || rac.STEP_LINE != 0) {
		if (check_range(&rac.LINE_START, &rac.LINE_END, rac.total_lines, RANGE_NOT_ADJUST) < 1) 
			return;
	}
	else
		if (check_range(&rac.LINE_START, &rac.LINE_END, rac.total_lines, RANGE_ADJUST) < 1) 
			return;

	/* define STEP_LINE, if it is not defined */
	if (rac.STEP_LINE == 0)
		rac.STEP_LINE = rac.LINE_START < rac.LINE_END ? 1 : -1;

#ifdef DEBUG2
	printf("--- Definitive Line Range = from %li to %li\n",rac.LINE_START,rac.LINE_END);
	printf("--- STEP_LINE = %li\n",rac.STEP_LINE);
	printf("--- col start %s end %s\n",col_cmd_start, col_cmd_end);
#endif

	for (i = rac.LINE_START; i != rac.LINE_END + rac.STEP_LINE; i += rac.STEP_LINE) {
		/* as for is i != LINE_END not < or > ... and we can have a STEP_LINE
		 * greater or less than 1. 
		 * so it could try to access a line away from the limits
		 * but we can permit this access 8=) */
		if (rac.LINE_START < rac.LINE_END && i > rac.LINE_END) 	break;
		if (rac.LINE_START > rac.LINE_END && i < rac.LINE_END) 	break;

		if (i == rac.LINE_START && i == rac.LINE_END) {
			col_start = col_cmd_start == NULL ? 1 : atoi(col_cmd_start);
			col_end = col_cmd_end == NULL ? -1 : atoi(col_cmd_end);
			if (check_range(&col_start, &col_end, rac.num_chars_line[i-1]-1, RANGE_ADJUST) < 1)
				break;
			locate_line(i);
			temp_line = get_line(i);
			for (j=col_start-1;j<col_end;j++)
				printf("%c",temp_line[j]);
			printf("\n");
		}
		else if (i == rac.LINE_START) {
			col_start = col_cmd_start == NULL ? 1 : atoi(col_cmd_start);
			/* num_chars_line[i-1] - 1
			 * less 1 because we can not count new line. */
			if (check_range(&col_start, &temp_col, rac.num_chars_line[i-1]-1, RANGE_NOT_ADJUST) < 1)
				continue;
			locate_line(i);
			temp_line = get_line(i);
			for (j=col_start-1;j<rac.num_chars_line[i-1];j++)
				printf("%c",temp_line[j]);
		}
		else if (i == rac.LINE_END) {
			col_end = col_cmd_end == NULL ? -1 : atoi(col_cmd_end);
			/* num_chars less 1, we can not count new line. */
			if (check_range(&temp_col, &col_end, rac.num_chars_line[i-1]-1, RANGE_NOT_ADJUST) < 1)
				continue;
			locate_line(i);
			temp_line = get_line(i);
			for (j=0;j<col_end;j++)
				printf("%c",temp_line[j]);
			printf("\n");
		}
		else {
			locate_line(i);
			temp_line = get_line(i);
			printf("%s",temp_line);
		}
		free(temp_line);
	}
}

/* separates the column command in start and end 
 * it do not analise the address, put in memory only the columns command
 * ex: col_cmd_start = 3:4,4,-1 col_cmd_end = 3
 */
void get_col_chunk(char *command, char **col_cmd_start, char **col_cmd_end)
{
	char c, *temp_col, *temp_command = command;
	int i=0, line_key = 0;

	temp_col = xmalloc(strlen(command)+1, "get_col_chunk alloc p1.");
	temp_col[0] = '\0';
	strncpy(&c,command,1);
	while (c) {
		if (c != '[' && i == 1) strncat(temp_col,command,1);
		else if (c == ':' && i == 0) line_key = 1; 	
		if (c == ']') {
			i = 0;
			/* erase the last char, ie, ']' */
			temp_col[strlen(temp_col)-1] = '\0';
			/* is it a start column command */
			if (line_key == 0) {
				*col_cmd_start = xmalloc(strlen(temp_col)+1,
						"get_col_chunk: alloc col_cmd_start.");
				strncpy(*col_cmd_start,temp_col,strlen(temp_col)+1);
			}
			/* it is a end column command */
			else {
				*col_cmd_end = xmalloc(strlen(temp_col)+1,
						"get_col_chunk: alloc col_cmd_end.");
				strncpy(*col_cmd_end,temp_col,strlen(temp_col)+1);
			}
			temp_col[0] = '\0';
		}
		if (c == '[') i = 1;
		command++;
		strncpy(&c,command,1);
	}
	/* do not have line range. ex: '[x,y]' address, so col_cmd_start 
	 * and col_cmd_end are equal */
	if (*col_cmd_start != NULL && line_key == 0) {
		*col_cmd_end = xmalloc(strlen(temp_col)+1,
				"get_col_chunk: alloc col_cmd_end.");
		strncpy(*col_cmd_end,*col_cmd_start,strlen(*col_cmd_start)+1);
	}
	command = temp_command;
#ifdef DEBUG2
	printf("--- col_cmd_start = %s\n",*col_cmd_start);
	printf("--- col_cmd_end   = %s\n",*col_cmd_end);
#endif
	free(temp_col);
}

/* execute the comands */
void exec_cmd(char *argv[])
{
	int i, k, type;
	char **command=NULL;
	/* it has the chunk of command that 
	 * has the columns address */
	char *col_cmd_start, *col_cmd_end;

	/* get the number of independent commands (separated by ';') */
	k = read_cmd(&command, convert_strings(argv[1]));

	/* it executes independent commands */
	for (i=0;i<k;i++) {
#ifdef DEBUG2
		printf("--- Command = %s\n",command[i]);
#endif
		col_cmd_start = col_cmd_end = NULL;
		rac.LINE_START = rac.LINE_END = rac.STEP_LINE = 0;

		get_line_range(command[i]);
		get_col_chunk(command[i], &col_cmd_start, &col_cmd_end);
		type = get_command_type(command[i]);
		if (type == TYPE_LINE_RANGE)
			exec_line_range(command[i], col_cmd_start, col_cmd_end);
		else if (type == TYPE_COL_RANGE)
			exec_col_range(command[i], col_cmd_start, col_cmd_end);

		if (col_cmd_start != NULL) free(col_cmd_start);
		if (col_cmd_end   != NULL) free(col_cmd_end);
		free(command[i]);
	}
	free(command);
}



/* convert the string addresses to line numbers addresses
 * in the command line
 */
char *convert_strings(char *line)
{
	char *conv_line, prev_char = 0;
	struct address_string *str = NULL;
	char col_mode = 0, str_mode = 0, pos_mode = 0;
	int i, j, l, line_size, str_start = 0, flag_col = 0, range = 0;

	line_size = strlen(line) + 1;
	conv_line = xmalloc(line_size, "convert_strings alloc conv_line.");
	line[line_size] = '\0';

	for (i = 0, j = 0; line[i] || pos_mode; i++) {
		/* in column mode, inside [ ] */
		if (col_mode) {
			if (str_mode) {
				if (line[i] == '\\' && line[i+1]) 
					i++;
				else if (line[i] == '=') 
					str_mode = 0;
			}
			else {
				if (line[i] == '=') 
					str_mode = 1;
				else if (line[i] == ']') {
					prev_char = line[i];
					col_mode = 0;
				}
			}
			conv_line[j++] = line[i];
		}
		/* in string mode, inside = = */
		else if (str_mode) {
			if (line[i] == '\\' && line[i+1])
				i++;
			else if (line[i] == '=') {
				line[i] = '\0';
				str = rac.strings;
				while (str) {
					if (!strcmp(&line[str_start], str->string)) 
						break;
					str = str->next;
				}
				if (!str) 
					error("convert_strings", "unexpected error searching string.");
				if (str->num_occurs <= 0) {
					if (prev_char == ':'){
						for (l=strlen(conv_line); l>=0 && conv_line[l] != ';'; l--)
							conv_line[l]='\0';
						j = l;
						prev_char = '\0';
					}
					break;
				}
				line[i] = '=';
				str_mode = 0;
				pos_mode = 1;
			}
		}
		/* in pos string mode, just after = = */
		else if (pos_mode) {
			char num_str[NUM_STR_SIZE], str_in_col = 0;
			int str_end, col_start = 0, col_end = 0;
			long int k, nl;
			regex_t re;
			regmatch_t match[7];	/* this size depends on number of () in ER_STR_MODIFIERS */

			if (regcomp(&re , ER_STR_MODIFIERS, REG_EXTENDED) != 0)
				error("convert_strings", "regcomp");
			regexec(&re, &line[i], 7, match, 0);
			/*
			match groups:
			0 = entire string matched
			1 = count occurrence
			2 = count negative occurrence
			3 = number (count occurrence)
			4 = count ahead or back
			5 = number (count ahead)
			6 = step or negative step
			*/

			/* for while, step without count occurrence is forbidden */
			if (match[6].rm_so != -1 && match[1].rm_so == -1)
				error("", "A ocorrncia inicial deve ser especificada.");
			/* check line after modifiers */
			str_end = i;
			i += match[0].rm_eo;
			if (line[i] == '[') {
				/* column without count occurrence set to all occurrence */
				if (match[1].rm_so == -1)
					flag_col = 1;
				col_start = i;
				for (i += 1; line[i]; i++) {
					if (str_in_col) {
						if (line[i] == '\\' && line[i+1])
							i++;
						else if (line[i] == '=')
							str_in_col = 0;
					}
					else {
						if (line[i] == '=') str_in_col = 1;
						else if (line[i] == ']') {
							col_end = i + 1;
							break;
						}
					}
				}
				for (i += 1; line[i] == ' ' || line[i] == '\t'; i++);
			}
			switch (line[i]) {
			case ':':
				/* for while, range without count occurrence is forbidden */
				if (match[1].rm_so == -1 || prev_char == ':')
					error("", "A ocorrncia deve ser especificada.");
				range = 1;
				break;
			case ';':
			case '\0':
				break;
			default:
				error("", "Invalid addressing.");
				break;
			}
			/* count ahead (or back) modifier */
			if (match[4].rm_so != -1) {
				if (match[5].rm_eo - match[5].rm_so > NUM_STR_SIZE)
					strncpy(num_str, &line[str_end + match[5].rm_so], NUM_STR_SIZE);
				else 
					strncpy(num_str, &line[str_end + match[5].rm_so], match[5].rm_eo - match[5].rm_so);
				nl = atol(num_str);
				if (nl > LONG_MAX) 
					error("", "Address overflow LONG_MAX.");
				if (line[str_end + match[4].rm_so] == '-') 
					nl = -nl;
			}
			else
				nl = 0;

			/* count occurrence (or negative occurrence) modifier */
			if (match[1].rm_so != -1) {
				if (match[3].rm_eo - match[3].rm_so > NUM_STR_SIZE)
					strncpy(num_str, &line[str_end + match[3].rm_so], NUM_STR_SIZE);
				else 
					strncpy(num_str, &line[str_end + match[3].rm_so], match[3].rm_eo - match[3].rm_so);
				k = atol(num_str);
				if (match[2].rm_eo - match[2].rm_so)
					k = str->num_occurs - k;
				else 
					k--;
				if (k < 0)
					nl = 0;
				else if (k >= str->num_occurs) 
					nl = rac.total_lines + 1;
				else {
					if (nl > 0 && nl + str->occurs[k] < 0)
						error("", "Address overflow LONG_MAX.");
					nl += str->occurs[k];
					if (nl < 0)
						nl = 0;
				}
				if (k < 0 && str->num_occurs ==0) nl=1;
				if (j + NUM_STR_SIZE > line_size) {
					line_size += NUM_STR_SIZE;
					conv_line = realloc(conv_line, line_size * sizeof(long int));
					if (!conv_line)
						error("convert_strings", "realloc of conv_line."); 
				}
				j += sprintf(&conv_line[j], "%ld", nl);
			}
			else {
				if (range == 1)
					error("", "A ocorrncia deve ser especificada nos dois lados do trecho.");
				/* all occurrences of string */
				for (k = 0; k < str->num_occurs; k++) {
					if (j + NUM_STR_SIZE > line_size) {
						line_size += NUM_STR_SIZE;
						conv_line = realloc(conv_line, line_size * sizeof(long int));
						if (!conv_line)
							error("convert_strings", "realloc of conv_line."); 
					}
					if (nl > 0 && nl + str->occurs[k] < 0) 
						error("", "Address overflow LONG_MAX.");
					if (nl + str->occurs[k] < 0) 
						conv_line[j++] = '0';
					else 
						j += sprintf(&conv_line[j], "%ld", nl + str->occurs[k]);
					/* specific column for all occurrences */
					if (col_start && flag_col == 1) {
						if (j + col_end - col_start + 2 > line_size) {
							line_size = j + col_end - col_start + 2;
							conv_line = realloc(conv_line, line_size * sizeof(long int));
							if (!conv_line)
								error("convert_strings", "realloc of conv_line."); 
						}
						strncpy(&conv_line[j], &line[col_start], col_end - col_start);
						j += col_end - col_start;
					}
					if (k + 1 != str->num_occurs) 
						conv_line[j++] = ';';
				}
			}
			/* step (or negative step) modifier */
			if (match[6].rm_so != -1) {
				if (j + match[6].rm_eo - match[6].rm_so + 2 > line_size) {
					line_size = j + match[6].rm_eo - match[6].rm_so + 2;
					conv_line = realloc(conv_line, line_size * sizeof(long int));
					if (!conv_line) 
						error("convert_strings", "realloc of conv_line."); 
				}
				strncpy(&conv_line[j], &line[str_end + match[6].rm_so], match[6].rm_eo - match[6].rm_so);
				j += match[6].rm_eo - match[6].rm_so;
			}
			/* column addressing complement */
			if (col_start && flag_col == 0) {
				if (j + col_end - col_start + 2 > line_size) {
					line_size = j + col_end - col_start + 2;
					conv_line = realloc(conv_line, line_size * sizeof(long int));
					if (!conv_line)
						error("convert_strings", "realloc of conv_line."); 
				}
				strncpy(&conv_line[j], &line[col_start], col_end - col_start);
				j += col_end - col_start;
			}
			prev_char = line[i];
			conv_line[j++] = line[i];
			pos_mode = 0;
			flag_col = 0;
		}
		/* out of any mode */
		else {
			if (line[i] == '=') {
				if (prev_char && prev_char != ';' && prev_char != ':')
					error("", "Invalid a addressing.");
				str_mode = 1;
				str_start = i + 1;
			}
			else {
				switch (line[i]) {
				case '[': col_mode = 1; break;
				case ' ':
				case '\t': break;
				default: prev_char = line[i];
				}
				conv_line[j++] = line[i];
			}
		}
	}
	conv_line[j] = '\0';
	return conv_line;
}


